from __future__ import annotations
from typing import Any, List, Sequence, Optional

import torch

from pulser.backend import State, Operator
from emu_base import DEVICE_COUNT
from emu_mps.algebra import add_factors, scale_factors, zip_right
from pulser.backend.operator import FullOp, QuditOp
from emu_mps.mps import MPS, DEFAULT_MAX_BOND_DIM, DEFAULT_PRECISION
from emu_mps.utils import new_left_bath, assign_devices

dtype = torch.complex128


class MPO(Operator[complex, torch.Tensor, MPS]):
    """
    Matrix Product Operator.

    Each tensor has 4 dimensions ordered as such: (left bond, output, input, right bond).

    Args:
        factors: the tensors making up the MPO
    """

    def __init__(
        self, factors: List[torch.Tensor], /, num_gpus_to_use: Optional[int] = None
    ):
        self.factors = factors
        self.num_sites = len(factors)
        if not self.num_sites > 1:
            raise ValueError("For 1 qubit states, do state vector")
        if factors[0].shape[0] != 1 or factors[-1].shape[-1] != 1:
            raise ValueError(
                "The dimension of the left (right) link of the first (last) "
                "tensor should be 1"
            )
        assert all(
            factors[i - 1].shape[-1] == factors[i].shape[0]
            for i in range(1, self.num_sites)
        )

        if num_gpus_to_use is not None:
            assign_devices(self.factors, min(DEVICE_COUNT, num_gpus_to_use))

    def __repr__(self) -> str:
        return "[" + ", ".join(map(repr, self.factors)) + "]"

    def apply_to(self, other: MPS) -> MPS:
        """
        Applies this MPO to the given MPS.
        The returned MPS is:

            - othogonal on the first site
            - truncated up to `other.precision`
            - distributed on the same devices of `other`

        Args:
            other: the state to apply this operator to

        Returns:
            the resulting state
        """
        assert isinstance(other, MPS), "MPO can only be multiplied with MPS"
        factors = zip_right(
            self.factors,
            other.factors,
            precision=other.precision,
            max_bond_dim=other.max_bond_dim,
        )
        return MPS(factors, orthogonality_center=0, eigenstates=other.eigenstates)

    def __add__(self, other: MPO) -> MPO:
        """
        Returns the sum of two MPOs, computed with a direct algorithm.
        The result is currently not truncated

        Args:
            other: the other operator

        Returns:
            the summed operator
        """
        assert isinstance(other, MPO), "MPO can only be added to another MPO"
        sum_factors = add_factors(self.factors, other.factors)
        return MPO(sum_factors)

    def __rmul__(self, scalar: complex) -> MPO:
        """
        Multiply an MPO by scalar.
        Assumes the orthogonal centre is on the first factor.

        Args:
            scalar: the scale factor to multiply with

        Returns:
            the scaled MPO
        """
        factors = scale_factors(self.factors, scalar, which=0)
        return MPO(factors)

    def __matmul__(self, other: MPO) -> MPO:
        """
        Compose two operators. The ordering is that
        self is applied after other.

        Args:
            other: the operator to compose with self

        Returns:
            the composed operator
        """
        assert isinstance(other, MPO), "MPO can only be applied to another MPO"
        factors = zip_right(
            self.factors,
            other.factors,
            precision=DEFAULT_PRECISION,
            max_bond_dim=DEFAULT_MAX_BOND_DIM,
        )
        return MPO(factors)

    def expect(self, state: State) -> torch.Tensor:
        """
        Compute the expectation value of self on the given state.

        Args:
            state: the state with which to compute

        Returns:
            the expectation
        """
        assert isinstance(
            state, MPS
        ), "currently, only expectation values of MPSs are \
        supported"
        acc = torch.ones(
            1, 1, 1, dtype=state.factors[0].dtype, device=state.factors[0].device
        )
        n = len(self.factors) - 1
        for i in range(n):
            acc = new_left_bath(acc, state.factors[i], self.factors[i]).to(
                state.factors[i + 1].device
            )
        acc = new_left_bath(acc, state.factors[n], self.factors[n])
        return acc.view(1)[0].cpu()

    @classmethod
    def _from_operator_repr(
        cls,
        *,
        eigenstates: Sequence[str],
        n_qudits: int,
        operations: FullOp[complex],
        **kwargs: Any,
    ) -> tuple[MPO, FullOp[complex]]:
        """
        See the base class

        Args:
            basis: the eigenstates in the basis to use e.g. ('r', 'g')
            nqubits: how many qubits there are in the state
            operations: which bitstrings make up the state with what weight
            operators: additional symbols to be used in operations

        Returns:
            the operator in MPO form.
        """

        basis = set(eigenstates)
        dim = len(basis)
        operators_with_tensors: dict[str, torch.Tensor | QuditOp]
        if basis == {"r", "g"}:
            # operators_with_tensors will now contain the basis for single
            # qubit ops, and potentially user defined strings in terms of these
            operators_with_tensors = {
                "gg": torch.tensor([[1.0, 0.0], [0.0, 0.0]], dtype=dtype).view(
                    1, dim, dim, 1
                ),
                "rg": torch.tensor([[0.0, 0.0], [1.0, 0.0]], dtype=dtype).view(
                    1, 2, 2, 1
                ),
                "gr": torch.tensor([[0.0, 1.0], [0.0, 0.0]], dtype=dtype).view(
                    1, 2, 2, 1
                ),
                "rr": torch.tensor([[0.0, 0.0], [0.0, 1.0]], dtype=dtype).view(
                    1, dim, dim, 1
                ),
            }
        elif basis == {"0", "1"}:
            # operators_with_tensors will now contain the basis for single
            # qubit ops, and potentially user defined strings in terms of these
            operators_with_tensors = {
                "00": torch.tensor([[1.0, 0.0], [0.0, 0.0]], dtype=dtype).view(
                    1, dim, dim, 1
                ),
                "10": torch.tensor([[0.0, 0.0], [1.0, 0.0]], dtype=dtype).view(
                    1, 2, 2, 1
                ),
                "01": torch.tensor([[0.0, 1.0], [0.0, 0.0]], dtype=dtype).view(
                    1, 2, 2, 1
                ),
                "11": torch.tensor([[0.0, 0.0], [0.0, 1.0]], dtype=dtype).view(
                    1, dim, dim, 1
                ),
            }
        elif basis == {"r", "g", "x"}:
            # operators_with_tensors will now contain the basis for single
            # qubit ops, and potentially user defined strings in terms of these
            operators_with_tensors = {
                "gg": torch.tensor(
                    [[1.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]], dtype=dtype
                ).view(1, dim, dim, 1),
                "gr": torch.tensor(
                    [[0.0, 1.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]], dtype=dtype
                ).view(1, dim, dim, 1),
                "rg": torch.tensor(
                    [[0.0, 0.0, 0.0], [1.0, 0.0, 0.0], [0.0, 0.0, 0.0]], dtype=dtype
                ).view(1, dim, dim, 1),
                "rr": torch.tensor(
                    [[0.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]], dtype=dtype
                ).view(1, dim, dim, 1),
                "xx": torch.tensor(
                    [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 1.0]], dtype=dtype
                ).view(1, dim, dim, 1),
                "xg": torch.tensor(
                    [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [1.0, 0.0, 0.0]], dtype=dtype
                ).view(1, dim, dim, 1),
                "xr": torch.tensor(
                    [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 1.0, 0.0]], dtype=dtype
                ).view(1, dim, dim, 1),
                "gx": torch.tensor(
                    [[0.0, 0.0, 1.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]], dtype=dtype
                ).view(1, dim, dim, 1),
                "rx": torch.tensor(
                    [[0.0, 0.0, 0.0], [0.0, 0.0, 1.0], [0.0, 0.0, 0.0]], dtype=dtype
                ).view(1, dim, dim, 1),
            }

        else:
            raise ValueError("Unsupported basis provided")

        mpos = []
        for coeff, tensorop in operations:
            # this function will recurse through the operators_with_tensors,
            # and replace any definitions
            # in terms of strings by the computed tensor
            def replace_operator_string(op: QuditOp | torch.Tensor) -> torch.Tensor:
                if isinstance(op, torch.Tensor):
                    return op

                result = torch.zeros(1, dim, dim, 1, dtype=dtype)
                for opstr, coeff in op.items():
                    tensor = replace_operator_string(operators_with_tensors[opstr])
                    operators_with_tensors[opstr] = tensor
                    result += tensor * coeff
                return result

            factors = [torch.eye(dim, dim, dtype=dtype).view(1, dim, dim, 1)] * n_qudits

            for op in tensorop:
                factor = replace_operator_string(op[0])
                for target_qubit in op[1]:
                    factors[target_qubit] = factor

            mpos.append(coeff * cls(factors, **kwargs))
        return sum(mpos[1:], start=mpos[0]), operations  # type: ignore[no-any-return]



from pulser.backend import EmulatorBackend, Results
from emu_mps.mps_config import MPSConfig
from emu_base import init_logging
from emu_mps.mps_backend_impl import create_impl, MPSBackendImpl
import pickle
import os
import time
import pathlib


class MPSBackend(EmulatorBackend):
    """
    A backend for emulating Pulser sequences using Matrix Product States (MPS),
    aka tensor trains.
    """

    default_config = MPSConfig()

    @staticmethod
    def resume(autosave_file: str | pathlib.Path) -> Results:
        """
        Resume simulation from autosave file.
        Only resume simulations from data you trust!
        Unpickling of untrusted data is not safe.
        """
        if isinstance(autosave_file, str):
            autosave_file = pathlib.Path(autosave_file)

        if not autosave_file.is_file():
            raise ValueError(f"Not a file: {autosave_file}")

        with open(autosave_file, "rb") as f:
            impl: MPSBackendImpl = pickle.load(f)

        impl.autosave_file = autosave_file
        impl.last_save_time = time.time()
        init_logging(impl.config.log_level, impl.config.log_file)

        impl.config.logger.warning(
            f"Resuming simulation from file {autosave_file}\n"
            f"Saving simulation state every {impl.config.autosave_dt} seconds"
        )

        return MPSBackend._run(impl)

    def run(self) -> Results:
        """
        Emulates the given sequence.

        Returns:
            the simulation results
        """
        assert isinstance(self._config, MPSConfig)

        impl = create_impl(self._sequence, self._config)
        impl.init()  # This is separate from the constructor for testing purposes.

        results = self._run(impl)

        return impl.permute_results(results, self._config.optimize_qubit_ordering)

    @staticmethod
    def _run(impl: MPSBackendImpl) -> Results:
        while not impl.is_finished():
            impl.progress()

        if impl.autosave_file.is_file():
            os.remove(impl.autosave_file)

        return impl.results



from typing import Any, ClassVar
from types import MethodType

import copy

from emu_base import init_logging
from emu_mps.mps import MPS, DEFAULT_MAX_BOND_DIM, DEFAULT_PRECISION
from emu_mps.mpo import MPO
from emu_mps.solver import Solver
from emu_mps.custom_callback_implementations import (
    energy_mps_impl,
    energy_second_moment_mps_impl,
    energy_variance_mps_impl,
    correlation_matrix_mps_impl,
    qubit_occupation_mps_impl,
)
from pulser.backend import (
    Occupation,
    CorrelationMatrix,
    Energy,
    EnergySecondMoment,
    EnergyVariance,
    BitStrings,
    EmulationConfig,
)
import logging
import pathlib


class MPSConfig(EmulationConfig):
    """
    The configuration of the emu-mps MPSBackend. The kwargs passed to this class
    are passed on to the base class.
    See the API for that class for a list of available options.

    Args:
        dt: The timestep size that the solver uses. Note that observables are
            only calculated if the evaluation_times are divisible by dt.
        precision: Up to what precision the state is truncated.
            Defaults to `1e-5`.
        max_bond_dim: The maximum bond dimension that the state is allowed to have.
            Defaults to `1024`.
        max_krylov_dim:
            The size of the krylov subspace that the Lanczos algorithm maximally builds
        extra_krylov_tolerance:
            The Lanczos algorithm uses this*precision as the convergence tolerance
        num_gpus_to_use: number of GPUs to be used in a given simulation.
            - if it is set to a number `n > 0`, the state will be distributed across `n` GPUs.
            - if it is set to `n = 0`, the entire simulation runs on the CPU.
            - if it is `None` (the default value), the backend internally chooses the number of GPUs
            based on the hardware availability during runtime.
        As shown in the benchmarks, using multiple GPUs might
            alleviate memory pressure per GPU, but the runtime should be similar.
        optimize_qubit_ordering: Optimize the register ordering. Improves performance and
            accuracy, but disables certain features.
        interaction_cutoff: Set interaction coefficients below this value to `0`.
            Potentially improves runtime and memory consumption.
        log_level: How much to log. Set to `logging.WARN` to get rid of the timestep info.
        log_file: If specified, log to this file rather than stout.
        autosave_prefix: filename prefix for autosaving simulation state to file
        autosave_dt: Minimum time interval in seconds between two autosaves.
            Saving the simulation state is only possible at specific times,
            therefore this interval is only a lower bound.
        solver: Chooses the solver algorithm to run a sequence.
            Two options are currently available:
            ``TDVP``, which performs ordinary time evolution,
            and ``DMRG``, which adiabatically follows the ground state
            of a given adiabatic pulse.
        kwargs: Arguments that are passed to the base class

    Examples:
        >>> num_gpus_to_use = 2 #use 2 gpus if available, otherwise 1 or cpu
        >>> dt = 1 #this will impact the runtime
        >>> precision = 1e-6 #smaller dt requires better precision, generally
        >>> MPSConfig(num_gpus_to_use=num_gpus_to_use, dt=dt, precision=precision,
        >>>     with_modulation=True) #the last arg is taken from the base class
    """

    # Whether to warn if unexpected kwargs are received
    _enforce_expected_kwargs: ClassVar[bool] = True
    _state_type = MPS
    _operator_type = MPO

    def __init__(
        self,
        *,
        dt: int = 10,
        precision: float = DEFAULT_PRECISION,
        max_bond_dim: int = DEFAULT_MAX_BOND_DIM,
        max_krylov_dim: int = 100,
        extra_krylov_tolerance: float = 1e-3,
        num_gpus_to_use: int | None = None,
        optimize_qubit_ordering: bool = True,
        interaction_cutoff: float = 0.0,
        log_level: int = logging.INFO,
        log_file: pathlib.Path | None = None,
        autosave_prefix: str = "emu_mps_save_",
        autosave_dt: int = 600,  # 10 minutes
        solver: Solver = Solver.TDVP,
        **kwargs: Any,
    ):
        kwargs.setdefault("observables", [BitStrings(evaluation_times=[1.0])])
        super().__init__(
            dt=dt,
            precision=precision,
            max_bond_dim=max_bond_dim,
            max_krylov_dim=max_krylov_dim,
            extra_krylov_tolerance=extra_krylov_tolerance,
            num_gpus_to_use=num_gpus_to_use,
            optimize_qubit_ordering=optimize_qubit_ordering,
            interaction_cutoff=interaction_cutoff,
            log_level=log_level,
            log_file=log_file,
            autosave_prefix=autosave_prefix,
            autosave_dt=autosave_dt,
            solver=solver,
            **kwargs,
        )

        MIN_AUTOSAVE_DT = 10
        assert (
            self.autosave_dt > MIN_AUTOSAVE_DT
        ), f"autosave_dt must be larger than {MIN_AUTOSAVE_DT} seconds"

        self.monkeypatch_observables()
        self.logger = init_logging(log_level, log_file)

        if (self.noise_model.runs != 1 and self.noise_model.runs is not None) or (
            self.noise_model.samples_per_run != 1
            and self.noise_model.samples_per_run is not None
        ):
            self.logger.warning(
                "Warning: The runs and samples_per_run values of the NoiseModel are ignored!"
            )
        self._backend_options[
            "optimize_qubit_ordering"
        ] &= self.check_permutable_observables()

    def _expected_kwargs(self) -> set[str]:
        return super()._expected_kwargs() | {
            "dt",
            "precision",
            "max_bond_dim",
            "max_krylov_dim",
            "extra_krylov_tolerance",
            "num_gpus_to_use",
            "optimize_qubit_ordering",
            "interaction_cutoff",
            "log_level",
            "log_file",
            "autosave_prefix",
            "autosave_dt",
            "solver",
        }

    def monkeypatch_observables(self) -> None:
        obs_list = []
        for _, obs in enumerate(self.observables):  # monkey patch
            obs_copy = copy.deepcopy(obs)
            if isinstance(obs, Occupation):
                obs_copy.apply = MethodType(  # type: ignore[method-assign]
                    qubit_occupation_mps_impl, obs_copy
                )
            elif isinstance(obs, EnergyVariance):
                obs_copy.apply = MethodType(  # type: ignore[method-assign]
                    energy_variance_mps_impl, obs_copy
                )
            elif isinstance(obs, EnergySecondMoment):
                obs_copy.apply = MethodType(  # type: ignore[method-assign]
                    energy_second_moment_mps_impl, obs_copy
                )
            elif isinstance(obs, CorrelationMatrix):
                obs_copy.apply = MethodType(  # type: ignore[method-assign]
                    correlation_matrix_mps_impl, obs_copy
                )
            elif isinstance(obs, Energy):
                obs_copy.apply = MethodType(  # type: ignore[method-assign]
                    energy_mps_impl, obs_copy
                )
            obs_list.append(obs_copy)
        self.observables = tuple(obs_list)

    def check_permutable_observables(self) -> bool:
        allowed_permutable_obs = set(
            [
                "bitstrings",
                "occupation",
                "correlation_matrix",
                "statistics",
                "energy",
                "energy_variance",
                "energy_second_moment",
            ]
        )

        actual_obs = set([obs._base_tag for obs in self.observables])
        not_allowed = actual_obs.difference(allowed_permutable_obs)
        if not_allowed:
            self.logger.warning(
                f"emu-mps allows only {allowed_permutable_obs} observables with"
                " `optimize_qubit_ordering = True`."
                f" you provided unsupported {not_allowed}"
                " using `optimize_qubit_ordering = False` instead."
            )
        return not_allowed == set()



"""
This file deals with creation of the MPO corresponding
to the Hamiltonian of a neutral atoms quantum processor.
"""

from abc import abstractmethod, ABC
from typing import Iterator, Literal, Union

from pulser.channels.base_channel import States
from emu_base import HamiltonianType
import torch
from emu_mps.mpo import MPO


dtype = torch.complex128


Eigenstate = Union[States, Literal["0", "1"]]


class Operators:
    id = torch.eye(2, dtype=dtype)
    id_3x3 = torch.eye(3, dtype=dtype)
    n = torch.tensor([[0.0, 0.0], [0.0, 1.0]], dtype=dtype)
    creation = torch.tensor([[0.0, 1.0], [0.0, 0.0]], dtype=dtype)
    sx = torch.tensor([[0.0, 0.5], [0.5, 0.0]], dtype=dtype)
    sy = torch.tensor([[0.0, -0.5j], [0.5j, 0.0]], dtype=dtype)


class HamiltonianMPOFactors(ABC):
    def __init__(self, interaction_matrix: torch.Tensor, dim: int = 2):
        assert interaction_matrix.ndim == 2, "interaction matrix is not a matrix"
        assert (
            interaction_matrix.shape[0] == interaction_matrix.shape[1]
        ), "interaction matrix is not square"
        self.dim = dim
        self.interaction_matrix = interaction_matrix.clone()
        self.interaction_matrix.fill_diagonal_(0.0)  # or assert
        self.qubit_count = self.interaction_matrix.shape[0]
        self.middle = self.qubit_count // 2
        self.identity = Operators.id if self.dim == 2 else Operators.id_3x3

    def __iter__(self) -> Iterator[torch.Tensor]:
        yield self.first_factor()

        for n in range(1, self.middle):
            yield self.left_factor(n)

        if self.qubit_count >= 3:
            yield self.middle_factor()

        for n in range(self.middle + 1, self.qubit_count - 1):
            yield self.right_factor(n)

        yield self.last_factor()

    @abstractmethod
    def first_factor(self) -> torch.Tensor:
        pass

    @abstractmethod
    def left_factor(self, n: int) -> torch.Tensor:
        pass

    @abstractmethod
    def middle_factor(self) -> torch.Tensor:
        pass

    @abstractmethod
    def right_factor(self, n: int) -> torch.Tensor:
        pass

    @abstractmethod
    def last_factor(self) -> torch.Tensor:
        pass


class RydbergHamiltonianMPOFactors(HamiltonianMPOFactors):
    def first_factor(self) -> torch.Tensor:
        has_right_interaction = self.interaction_matrix[0, 1:].any()
        fac = torch.zeros(
            1, self.dim, self.dim, 3 if has_right_interaction else 2, dtype=dtype
        )
        fac[0, :, :, 1] = self.identity
        if has_right_interaction:
            fac[0, :2, :2, 2] = Operators.n

        return fac

    def left_factor(self, n: int) -> torch.Tensor:
        has_right_interaction = self.interaction_matrix[n, n + 1 :].any()
        current_left_interactions = self.interaction_matrix[:n, n:].any(dim=1)
        left_interactions_to_keep = self.interaction_matrix[:n, n + 1 :].any(dim=1)

        fac = torch.zeros(
            int(current_left_interactions.sum().item() + 2),
            self.dim,
            self.dim,
            int(left_interactions_to_keep.sum().item() + int(has_right_interaction) + 2),
            dtype=dtype,
        )

        fac[0, :, :, 0] = self.identity
        fac[1, :, :, 1] = self.identity
        if has_right_interaction:
            fac[1, :2, :2, -1] = Operators.n

        fac[2:, :2, :2, 0] = (
            self.interaction_matrix[:n][current_left_interactions, n, None, None]
            * Operators.n
        )

        i = 2
        j = 2
        for current_left_interaction in current_left_interactions.nonzero().flatten():
            if left_interactions_to_keep[current_left_interaction]:
                fac[i, :, :, j] = self.identity
                j += 1
            i += 1
        return fac

    def middle_factor(self) -> torch.Tensor:
        n = self.middle
        current_left_interactions = self.interaction_matrix[:n, n:].any(dim=1)
        current_right_interactions = self.interaction_matrix[n + 1 :, : n + 1].any(dim=1)

        fac = torch.zeros(
            int(current_left_interactions.sum().item() + 2),
            self.dim,
            self.dim,
            int(current_right_interactions.sum().item() + 2),
            dtype=dtype,
        )

        fac[0, :, :, 0] = self.identity
        fac[1, :, :, 1] = self.identity

        fac[2:, :2, :2, 0] = (
            self.interaction_matrix[:n][current_left_interactions, n, None, None]
            * Operators.n
        )

        fac[1, :2, :2, 2:] = self.interaction_matrix[n + 1 :][
            None, None, current_right_interactions, n
        ] * Operators.n.unsqueeze(-1)

        fac[2:, :, :, 2:] = (
            self.interaction_matrix[:n, n + 1 :][current_left_interactions, :][
                :, None, None, current_right_interactions
            ]
            * self.identity[None, ..., None]
        )

        return fac

    def right_factor(self, n: int) -> torch.Tensor:
        has_left_interaction = self.interaction_matrix[n, :n].any()
        current_right_interactions = self.interaction_matrix[n + 1 :, : n + 1].any(dim=1)
        right_interactions_to_keep = self.interaction_matrix[n + 1 :, :n].any(dim=1)

        fac = torch.zeros(
            int(right_interactions_to_keep.sum().item() + int(has_left_interaction) + 2),
            self.dim,
            self.dim,
            int(current_right_interactions.sum().item() + 2),
            dtype=dtype,
        )

        fac[0, :, :, 0] = self.identity
        fac[1, :, :, 1] = self.identity
        if has_left_interaction:
            fac[2, :2, :2, 0] = Operators.n

        fac[1, :2, :2, 2:] = self.interaction_matrix[n + 1 :][
            None, None, current_right_interactions, n
        ] * Operators.n.unsqueeze(-1)

        i = 3 if has_left_interaction else 2
        j = 2
        for current_right_interaction in current_right_interactions.nonzero().flatten():
            if right_interactions_to_keep[current_right_interaction]:
                fac[i, :, :, j] = self.identity
                i += 1
            j += 1
        return fac

    def last_factor(self) -> torch.Tensor:
        has_left_interaction = self.interaction_matrix[-1, :-1].any()
        fac = torch.zeros(
            3 if has_left_interaction else 2, self.dim, self.dim, 1, dtype=dtype
        )
        fac[0, :, :, 0] = self.identity
        if has_left_interaction:
            if self.qubit_count >= 3:
                fac[2, :2, :2, 0] = Operators.n
            else:
                fac[2, :2, :2, 0] = self.interaction_matrix[0, 1] * Operators.n

        return fac


class XYHamiltonianMPOFactors(HamiltonianMPOFactors):
    def first_factor(self) -> torch.Tensor:
        has_right_interaction = self.interaction_matrix[0, 1:].any()
        fac = torch.zeros(
            1, self.dim, self.dim, 4 if has_right_interaction else 2, dtype=dtype
        )
        fac[0, :, :, 1] = self.identity
        if has_right_interaction:
            fac[0, :2, :2, 2] = Operators.creation
            fac[0, :2, :2, 3] = Operators.creation.T

        return fac

    def left_factor(self, n: int) -> torch.Tensor:
        has_right_interaction = self.interaction_matrix[n, n + 1 :].any()
        current_left_interactions = self.interaction_matrix[:n, n:].any(dim=1)
        left_interactions_to_keep = self.interaction_matrix[:n, n + 1 :].any(dim=1)

        fac = torch.zeros(
            int(2 * current_left_interactions.sum().item() + 2),
            self.dim,
            self.dim,
            int(
                2 * left_interactions_to_keep.sum().item()
                + 2 * int(has_right_interaction)
                + 2
            ),
            dtype=dtype,
        )

        fac[0, :, :, 0] = self.identity
        fac[1, :, :, 1] = self.identity
        if has_right_interaction:
            fac[1, :2, :2, -2] = Operators.creation
            fac[1, :2, :2, -1] = Operators.creation.T

        fac[2::2, :2, :2, 0] = (
            self.interaction_matrix[:n][current_left_interactions, n, None, None]
            * Operators.creation.T
        )
        fac[3::2, :2, :2, 0] = (
            self.interaction_matrix[:n][current_left_interactions, n, None, None]
            * Operators.creation
        )

        i = 2
        j = 2
        for current_left_interaction in current_left_interactions.nonzero().flatten():
            if left_interactions_to_keep[current_left_interaction]:
                fac[i, :, :, j] = self.identity
                fac[i + 1, :, :, j + 1] = self.identity
                j += 2
            i += 2
        return fac

    def middle_factor(self) -> torch.Tensor:
        n = self.middle
        current_left_interactions = self.interaction_matrix[:n, n:].any(dim=1)
        current_right_interactions = self.interaction_matrix[n + 1 :, : n + 1].any(dim=1)

        fac = torch.zeros(
            int(2 * current_left_interactions.sum().item() + 2),
            self.dim,
            self.dim,
            int(2 * current_right_interactions.sum().item() + 2),
            dtype=dtype,
        )

        fac[0, :, :, 0] = self.identity
        fac[1, :, :, 1] = self.identity

        fac[2::2, :2, :2, 0] = (
            self.interaction_matrix[:n][current_left_interactions, n, None, None]
            * Operators.creation.T
        )
        fac[3::2, :2, :2, 0] = (
            self.interaction_matrix[:n][current_left_interactions, n, None, None]
            * Operators.creation
        )

        fac[1, :2, :2, 2::2] = self.interaction_matrix[n + 1 :][
            None, None, current_right_interactions, n
        ] * Operators.creation.unsqueeze(-1)
        fac[1, :2, :2, 3::2] = self.interaction_matrix[n + 1 :][
            None, None, current_right_interactions, n
        ] * Operators.creation.T.unsqueeze(-1)

        fac[2::2, :, :, 2::2] = (
            self.interaction_matrix[:n, n + 1 :][current_left_interactions, :][
                :, None, None, current_right_interactions
            ]
            * self.identity[None, ..., None]
        )
        fac[3::2, :, :, 3::2] = (
            self.interaction_matrix[:n, n + 1 :][current_left_interactions, :][
                :, None, None, current_right_interactions
            ]
            * self.identity[None, ..., None]
        )

        return fac

    def right_factor(self, n: int) -> torch.Tensor:
        has_left_interaction = self.interaction_matrix[n, :n].any()
        current_right_interactions = self.interaction_matrix[n + 1 :, : n + 1].any(dim=1)
        right_interactions_to_keep = self.interaction_matrix[n + 1 :, :n].any(dim=1)

        fac = torch.zeros(
            int(
                2 * right_interactions_to_keep.sum().item()
                + 2 * int(has_left_interaction)
                + 2
            ),
            self.dim,
            self.dim,
            int(2 * current_right_interactions.sum().item() + 2),
            dtype=dtype,
        )

        fac[0, :, :, 0] = self.identity
        fac[1, :, :, 1] = self.identity
        if has_left_interaction:
            fac[2, :2, :2, 0] = Operators.creation.T
            fac[3, :2, :2, 0] = Operators.creation

        fac[1, :2, :2, 2::2] = self.interaction_matrix[n + 1 :][
            None, None, current_right_interactions, n
        ] * Operators.creation.unsqueeze(-1)
        fac[1, :2, :2, 3::2] = self.interaction_matrix[n + 1 :][
            None, None, current_right_interactions, n
        ] * Operators.creation.T.unsqueeze(-1)

        i = 4 if has_left_interaction else 2
        j = 2
        for current_right_interaction in current_right_interactions.nonzero().flatten():
            if right_interactions_to_keep[current_right_interaction]:
                fac[i, :, :, j] = self.identity
                fac[i + 1, :, :, j + 1] = self.identity
                i += 2
            j += 2
        return fac

    def last_factor(self) -> torch.Tensor:
        has_left_interaction = self.interaction_matrix[-1, :-1].any()
        fac = torch.zeros(
            4 if has_left_interaction else 2, self.dim, self.dim, 1, dtype=dtype
        )
        fac[0, :, :, 0] = self.identity
        if has_left_interaction:
            if self.qubit_count >= 3:
                fac[2, :2, :2, 0] = Operators.creation.T
                fac[3, :2, :2, 0] = Operators.creation
            else:
                fac[2, :2, :2, 0] = self.interaction_matrix[0, 1] * Operators.creation.T
                fac[3, :2, :2, 0] = self.interaction_matrix[0, 1] * Operators.creation

        return fac


def make_H(
    *,
    interaction_matrix: torch.Tensor,  # depends on Hamiltonian Type
    hamiltonian_type: HamiltonianType,
    dim: int = 2,
    num_gpus_to_use: int | None,
) -> MPO:
    r"""
    Constructs and returns a Matrix Product Operator (MPO) representing the
    neutral atoms Hamiltonian, parameterized by `omega`, `delta`, and `phi`.

    The Hamiltonian H is given by:
    H = ∑ⱼΩⱼ[cos(ϕⱼ)σˣⱼ + sin(ϕⱼ)σʸⱼ] - ∑ⱼΔⱼnⱼ + ∑ᵢ﹥ⱼC⁶/rᵢⱼ⁶ nᵢnⱼ

    If noise is considered, the Hamiltonian includes an additional term to
    support the Monte Carlo WaveFunction algorithm:
    H = ∑ⱼΩⱼ[cos(ϕⱼ)σˣⱼ + sin(ϕⱼ)σʸⱼ] - ∑ⱼΔⱼnⱼ + ∑ᵢ﹥ⱼC⁶/rᵢⱼ⁶ nᵢnⱼ - 0.5i∑ₘ ∑ᵤ Lₘᵘ⁺ Lₘᵘ
    where Lₘᵘ are the Lindblad operators representing the noise,
    m for noise channel and u for the number of atoms

    make_H constructs an MPO of the appropriate size, but the single qubit
    terms are left at zero.
    To fill in the appropriate values, call update_H

    Args:
        interaction_matrix (torch.Tensor): The interaction matrix describing
        the interactions between qubits.
        hamiltonian_type: whether to use XY or Rydberg interation
        dim: dimension of the basis (2 or 3)
        num_gpus_to_use (int): how many gpus to put the Hamiltonian on.
        See utils.assign_devices
    Returns:
        MPO: A Matrix Product Operator (MPO) representing the specified
        Hamiltonian.

    Note:
    For more information about the Hamiltonian and its usage, refer to the
    [Pulser documentation](https://pulser.readthedocs.io/en/stable/conventions.html#hamiltonians).

    """

    if hamiltonian_type == HamiltonianType.Rydberg:
        return MPO(
            list(RydbergHamiltonianMPOFactors(interaction_matrix, dim=dim)),
            num_gpus_to_use=num_gpus_to_use,
        )

    if hamiltonian_type == HamiltonianType.XY:
        return MPO(
            list(XYHamiltonianMPOFactors(interaction_matrix, dim=dim)),
            num_gpus_to_use=num_gpus_to_use,
        )


def update_H(
    hamiltonian: MPO,
    omega: torch.Tensor,
    delta: torch.Tensor,
    phi: torch.Tensor,
    noise: torch.Tensor,
) -> None:
    """
    The single qubit operators in the Hamiltonian,
    corresponding to the omega, delta, phi parameters and the aggregated
    Lindblad operators have a well-determined position in the factors of
    the Hamiltonian.
    This function updates this part of the factors to update the
    Hamiltonian with new parameters without rebuilding the entire thing.
    See make_H for details about the Hamiltonian.

    This is an in-place operation, so this function returns nothing.

    Args:
        omega (torch.Tensor): Rabi frequency Ωⱼ for each qubit.
        delta (torch.Tensor): The detuning value Δⱼ for each qubit.
        phi (torch.Tensor): The phase ϕⱼ corresponding to each qubit.
        noise (torch.Tensor, optional): The single-qubit noise
        term -0.5i∑ⱼLⱼ†Lⱼ applied to all qubits.
        This can be computed using the `compute_noise_from_lindbladians`
        function.
        Defaults to a zero tensor.
    """

    assert noise.shape == (2, 2) or (3, 3)
    nqubits = omega.size(dim=0)

    a = torch.tensordot(omega * torch.cos(phi), Operators.sx, dims=0)
    c = torch.tensordot(delta, Operators.n, dims=0)
    b = torch.tensordot(omega * torch.sin(phi), Operators.sy, dims=0)

    factors = hamiltonian.factors

    single_qubit_terms = torch.stack(nqubits * [noise])

    single_qubit_terms[:, :2, :2] += a + b - c

    factors[0][0, :, :, 0] = single_qubit_terms[0]
    for i in range(1, nqubits):
        factors[i][1, :, :, 0] = single_qubit_terms[i]



from __future__ import annotations


import torch
import math
from emu_mps.utils import truncate_impl


def add_factors(
    left: list[torch.Tensor], right: list[torch.Tensor]
) -> list[torch.Tensor]:
    """
    Direct sum algorithm implementation to sum two tensor trains (MPS/MPO).
    It assumes the left and right bond are along the dimension 0 and -1 of each tensor.
    """
    num_sites = len(left)
    if num_sites != len(right):
        raise ValueError("Cannot sum two matrix products of different number of sites")

    new_tt = []
    for i, (core1, core2) in enumerate(zip(left, right)):
        core2 = core2.to(core1.device)
        if i == 0:
            core = torch.cat((core1, core2), dim=-1)  # concatenate along the right bond
        elif i == (num_sites - 1):
            core = torch.cat((core1, core2), dim=0)  # concatenate along the left bond
        else:
            pad_shape_1 = (core2.shape[0], *core1.shape[1:])
            padded_c1 = torch.cat(
                (
                    core1,
                    torch.zeros(pad_shape_1, device=core1.device, dtype=core1.dtype),
                ),
                dim=0,  # concatenate along the left bond
            )
            pad_shape_2 = (core1.shape[0], *core2.shape[1:])
            padded_c2 = torch.cat(
                (
                    torch.zeros(pad_shape_2, device=core1.device, dtype=core1.dtype),
                    core2,
                ),
                dim=0,  # concatenate along the left bond
            )
            core = torch.cat(
                (padded_c1, padded_c2), dim=-1
            )  # concatenate along the right bond
        new_tt.append(core)
    return new_tt


def scale_factors(
    factors: list[torch.Tensor], scalar: complex | torch.Tensor, *, which: int
) -> list[torch.Tensor]:
    """
    Returns a new list of factors where the tensor at the given index is scaled by `scalar`.
    """
    return [scalar * f if i == which else f for i, f in enumerate(factors)]


def zip_right_step(
    slider: torch.Tensor,
    top: torch.Tensor,
    bottom: torch.Tensor,
) -> tuple[torch.Tensor, torch.Tensor]:
    """
    Returns a new `MPS/O` factor of the result of the multiplication MPO @ MPS/O,
    and the updated slider, performing a single step of the
    [zip-up algorithm](https://tensornetwork.org/mps/algorithms/zip_up_mpo/).

    Args:
    - `slider`: utility tensor for the zip-up algorithm.
    - `top`: factor of the applied MPO.
    - `bottom`: factor of the MPS/O to which the MPO is being applied.

    First, moves all tensors to `bottom.device`.
    Second, it contracts `top` and then `bottom` to `slider`.
    The resulting tensor is then QR factorized into a
    new factor and the updated slider for the next zip step.

    Note:
    The method assumes that:
        - `top` is a valid MPO factor of shape
    (left_link_dim, out_site_dim, in_site_dim, right_link_dim).
        - `bottom` is a valid MPO/S factor
    """
    if slider.shape[1:] != (top.shape[0], bottom.shape[0]):
        msg = (
            f"Contracted dimensions between the slider, {slider.shape[1:]} on dims 1 and 2, "
            f"and the two factors, {(top.shape[0], bottom.shape[0])} on dim 0, need to match."
        )
        raise ValueError(msg)

    slider = slider.to(bottom.device)
    top = top.to(bottom.device)

    # merge top and bottom into slider
    slider = torch.tensordot(slider, top, dims=([1], [0]))
    slider = torch.tensordot(slider, bottom, dims=([3, 1], [1, 0]))

    if len(bottom.shape) == 4:  # MPO factor
        slider = slider.transpose(2, 3)

    # reshape slider as matrix
    left_inds = (slider.shape[0], *bottom.shape[1:-1])
    right_inds = (top.shape[-1], bottom.shape[-1])
    slider = slider.contiguous().view(math.prod(left_inds), math.prod(right_inds))

    L, slider = torch.linalg.qr(slider)

    # reshape slider to its original shape
    slider = slider.view((-1, *right_inds))
    # reshape left as MPS/O factor and
    return L.view(*left_inds, -1), slider


def zip_right(
    top_factors: list[torch.Tensor],
    bottom_factors: list[torch.Tensor],
    precision: float,
    max_bond_dim: int,
) -> list[torch.Tensor]:
    """
    Returns a new matrix product, resulting from applying `top` to `bottom`.
    The resulting factors are:
     - of the same order as `bottom` factors
     - on the same device of `bottom` factors
     - orthogonalized on the first element
     - truncated to `max_error`/`max_rank`

    Args:
    - `top`: MPO factors to be applied.
    - `bottom`: MPS/O factors to which the MPO factors are being applied.

    Note:
        Implements a general [zip-up](https://tensornetwork.org/mps/algorithms/zip_up_mpo/)
        algorithm for applying MPO factors to both MPO and MPS factors.
        A final truncation sweep, from right to left,
        moves back the orthogonal center to the first element.
    """
    if len(top_factors) != len(bottom_factors):
        raise ValueError("Cannot multiply two matrix products of different lengths.")

    slider = torch.ones(1, 1, 1, dtype=torch.complex128)
    new_factors = []
    for top, bottom in zip(top_factors, bottom_factors):
        res, slider = zip_right_step(slider, top, bottom)
        new_factors.append(res)
    new_factors[-1] @= slider[:, :, 0]

    truncate_impl(new_factors, precision=precision, max_bond_dim=max_bond_dim)

    return new_factors



import math
import os
import pathlib
import pickle
import random
import time
import typing
import uuid

from copy import deepcopy
from collections import Counter
from enum import Enum, auto
from types import MethodType
from typing import Any, Optional

import torch
from pulser import Sequence
from pulser.backend import EmulationConfig, Observable, Results, State

from emu_base import DEVICE_COUNT, PulserData, get_max_rss
from emu_base.math.brents_root_finding import BrentsRootFinder
from emu_base.utils import deallocate_tensor

from emu_mps.hamiltonian import make_H, update_H
from emu_mps.mpo import MPO
from emu_mps.mps import MPS
from emu_mps.mps_config import MPSConfig
from emu_base.jump_lindblad_operators import compute_noise_from_lindbladians
import emu_mps.optimatrix as optimat
from emu_mps.solver import Solver
from emu_mps.solver_utils import (
    evolve_pair,
    evolve_single,
    minimize_energy_pair,
    new_right_bath,
    right_baths,
)
from emu_mps.utils import (
    extended_mpo_factors,
    extended_mps_factors,
    get_extended_site_index,
    new_left_bath,
)

dtype = torch.complex128


class Statistics(Observable):
    def __init__(
        self,
        evaluation_times: typing.Sequence[float] | None,
        data: list[float],
        timestep_count: int,
    ):
        super().__init__(evaluation_times=evaluation_times)
        self.data = data
        self.timestep_count = timestep_count

    @property
    def _base_tag(self) -> str:
        return "statistics"

    def apply(
        self,
        *,
        config: EmulationConfig,
        state: State,
        **kwargs: Any,
    ) -> dict:
        """Calculates the observable to store in the Results."""
        assert isinstance(state, MPS)
        duration = self.data[-1]
        max_mem = get_max_rss(state.factors[0].is_cuda)

        config.logger.info(
            f"step = {len(self.data)}/{self.timestep_count}, "
            + f"χ = {state.get_max_bond_dim()}, "
            + f"|ψ| = {state.get_memory_footprint():.3f} MB, "
            + f"RSS = {max_mem:.3f} MB, "
            + f"Δt = {duration:.3f} s"
        )

        return {
            "max_bond_dimension": state.get_max_bond_dim(),
            "memory_footprint": state.get_memory_footprint(),
            "RSS": max_mem,
            "duration": duration,
        }


class SwipeDirection(Enum):
    LEFT_TO_RIGHT = auto()
    RIGHT_TO_LEFT = auto()


class MPSBackendImpl:
    current_time: float = (
        0.0  # While dt is an integer, noisy collapse can happen at non-integer times.
    )
    well_prepared_qubits_filter: Optional[torch.Tensor]
    hamiltonian: MPO
    state: MPS
    right_baths: list[torch.Tensor]
    sweep_index: int
    swipe_direction: SwipeDirection
    timestep_index: int
    target_time: float
    results: Results

    def __init__(self, mps_config: MPSConfig, pulser_data: PulserData):
        self.config = mps_config
        self.target_times = pulser_data.target_times
        self.target_time = self.target_times[1]
        self.pulser_data = pulser_data
        self.qubit_count = pulser_data.qubit_count
        assert self.qubit_count >= 2
        self.omega = pulser_data.omega
        self.delta = pulser_data.delta
        self.phi = pulser_data.phi
        self.timestep_count: int = self.omega.shape[0]
        self.has_lindblad_noise = pulser_data.has_lindblad_noise
        self.dim = pulser_data.dim
        self.lindblad_noise = torch.zeros(self.dim, self.dim, dtype=dtype)
        self.qubit_permutation = (
            optimat.minimize_bandwidth(pulser_data.full_interaction_matrix)
            if self.config.optimize_qubit_ordering
            else optimat.eye_permutation(self.qubit_count)
        )
        self.full_interaction_matrix = optimat.permute_tensor(
            pulser_data.full_interaction_matrix, self.qubit_permutation
        )

        self.masked_interaction_matrix = optimat.permute_tensor(
            pulser_data.masked_interaction_matrix, self.qubit_permutation
        )
        self.hamiltonian_type = pulser_data.hamiltonian_type
        self.slm_end_time = pulser_data.slm_end_time
        self.is_masked = self.slm_end_time > 0.0
        self.left_baths: list[torch.Tensor]
        self.time = time.time()
        self.swipe_direction = SwipeDirection.LEFT_TO_RIGHT
        self.sweep_index = 0
        self.timestep_index = 0
        self.results = Results(
            atom_order=optimat.permute_tuple(
                pulser_data.qubit_ids, self.qubit_permutation
            ),
            total_duration=self.target_times[-1],
        )
        self.statistics = Statistics(
            evaluation_times=[t / self.target_times[-1] for t in self.target_times],
            data=[],
            timestep_count=self.timestep_count,
        )
        self.autosave_file = self._get_autosave_filepath(self.config.autosave_prefix)
        self.config.logger.debug(
            f"""Will save simulation state to file "{self.autosave_file.name}"
            every {self.config.autosave_dt} seconds.\n"""
            f"""To resume: `MPSBackend().resume("{self.autosave_file}")`"""
        )
        self.last_save_time = time.time()
        requested_num_gpus = self.config.num_gpus_to_use

        if requested_num_gpus is None:
            requested_num_gpus = DEVICE_COUNT
        elif requested_num_gpus > DEVICE_COUNT:
            self.config.logger.warning(
                f"Requested to use {requested_num_gpus} GPU(s) "
                f"but only {DEVICE_COUNT if DEVICE_COUNT > 0 else 'cpu'} available"
            )
        self.resolved_num_gpus = requested_num_gpus

    def __getstate__(self) -> dict:
        d = self.__dict__.copy()
        cp = deepcopy(self.config)
        d["config"] = cp
        d["state"].config = cp
        for obs in cp.observables:
            obs.apply = MethodType(type(obs).apply, obs)  # type: ignore[method-assign]
        # mypy thinks the method below is an attribute, because of the __getattr__ override
        d["results"] = self.results._to_abstract_repr()  # type: ignore[operator]
        return d

    def __setstate__(self, d: dict) -> None:
        self.__dict__ = d
        self.results = Results._from_abstract_repr(d["results"])  # type: ignore [attr-defined]
        self.config.monkeypatch_observables()

    @staticmethod
    def _get_autosave_filepath(autosave_prefix: str) -> pathlib.Path:
        return pathlib.Path(os.getcwd()) / (autosave_prefix + str(uuid.uuid1()) + ".dat")

    def init_dark_qubits(self) -> None:
        # has_state_preparation_error
        if self.config.noise_model.state_prep_error > 0.0:
            bad_atoms = self.pulser_data.hamiltonian.bad_atoms
            self.well_prepared_qubits_filter = torch.logical_not(
                torch.tensor(list(bool(x) for x in bad_atoms.values()))
            )
        else:
            self.well_prepared_qubits_filter = None

        if self.well_prepared_qubits_filter is not None:
            self.qubit_count = sum(1 for x in self.well_prepared_qubits_filter if x)

            self.full_interaction_matrix = self.full_interaction_matrix[
                self.well_prepared_qubits_filter, :
            ][:, self.well_prepared_qubits_filter]
            self.masked_interaction_matrix = self.masked_interaction_matrix[
                self.well_prepared_qubits_filter, :
            ][:, self.well_prepared_qubits_filter]
            self.omega = self.omega[:, self.well_prepared_qubits_filter]
            self.delta = self.delta[:, self.well_prepared_qubits_filter]
            self.phi = self.phi[:, self.well_prepared_qubits_filter]

    def init_initial_state(self, initial_state: State | None = None) -> None:
        if initial_state is None:
            self.state = MPS.make(
                self.qubit_count,
                precision=self.config.precision,
                max_bond_dim=self.config.max_bond_dim,
                num_gpus_to_use=self.resolved_num_gpus,
            )
            return

        if self.well_prepared_qubits_filter is not None:
            raise NotImplementedError(
                "Specifying the initial state in the presence "
                "of state preparation errors is currently not implemented."
            )

        assert isinstance(initial_state, MPS)
        if not torch.equal(
            self.qubit_permutation, optimat.eye_permutation(self.qubit_count)
        ):
            # permute the initial state to match with permuted Hamiltonian
            abstr_repr = initial_state._to_abstract_repr()
            eigs = abstr_repr["eigenstates"]
            ampl = {
                optimat.permute_string(bstr, self.qubit_permutation): amp
                for bstr, amp in abstr_repr["amplitudes"].items()
            }
            initial_state = MPS.from_state_amplitudes(eigenstates=eigs, amplitudes=ampl)

        initial_state = MPS(
            # Deep copy of every tensor of the initial state.
            [f.detach().clone() for f in initial_state.factors],
            precision=self.config.precision,
            max_bond_dim=self.config.max_bond_dim,
            num_gpus_to_use=self.resolved_num_gpus,
            eigenstates=initial_state.eigenstates,
        )
        initial_state.truncate()
        initial_state *= 1 / initial_state.norm()
        self.state = initial_state
        self.state.orthogonalize(0)

    def init_hamiltonian(self) -> None:
        """
        Must be called AFTER init_dark_qubits otherwise,
        too many factors are put in the Hamiltonian
        """
        self.hamiltonian = make_H(
            interaction_matrix=(
                self.masked_interaction_matrix
                if self.is_masked
                else self.full_interaction_matrix
            ),
            hamiltonian_type=self.hamiltonian_type,
            num_gpus_to_use=self.resolved_num_gpus,
        )

        update_H(
            hamiltonian=self.hamiltonian,
            omega=self.omega[self.timestep_index, :],
            delta=self.delta[self.timestep_index, :],
            phi=self.phi[self.timestep_index, :],
            noise=self.lindblad_noise,
        )

    def init_baths(self) -> None:
        self.left_baths = [
            torch.ones(1, 1, 1, dtype=dtype, device=self.state.factors[0].device)
        ]
        self.right_baths = right_baths(self.state, self.hamiltonian, final_qubit=2)
        assert len(self.right_baths) == self.qubit_count - 1

    def get_current_right_bath(self) -> torch.Tensor:
        return self.right_baths[-1]

    def get_current_left_bath(self) -> torch.Tensor:
        return self.left_baths[-1]

    def init(self) -> None:
        self.init_dark_qubits()
        self.init_initial_state(self.config.initial_state)
        self.init_hamiltonian()
        self.init_baths()

    def is_finished(self) -> bool:
        return self.timestep_index >= self.timestep_count

    def _evolve(
        self, *indices: int, dt: float, orth_center_right: Optional[bool] = None
    ) -> None:
        """
        Time-evolve the state's tensors located at the given 1 or 2 indices by dt,
        using the baths stored in self.left_baths and self.right_baths.
        When 2 indices are given, they need to be consecutive.
        Updates the state's orthogonality center according to orth_center_right.
        """
        assert 1 <= len(indices) <= 2

        baths = (self.get_current_left_bath(), self.get_current_right_bath())

        if len(indices) == 1:
            assert orth_center_right is None
            (index,) = indices
            assert self.state.orthogonality_center == index

            self.state.factors[index] = evolve_single(
                state_factor=self.state.factors[index],
                ham_factor=self.hamiltonian.factors[index],
                baths=baths,
                dt=dt,
                config=self.config,
                is_hermitian=not self.has_lindblad_noise,
            )
        else:
            assert orth_center_right is not None
            l, r = indices
            assert r == l + 1, "Indices need to be consecutive"
            assert self.state.orthogonality_center in {l, r}, (
                "State needs to be orthogonalized" " on one of the evolved indices"
            )

            self.state.factors[l : r + 1] = evolve_pair(
                state_factors=self.state.factors[l : r + 1],
                ham_factors=self.hamiltonian.factors[l : r + 1],
                baths=baths,
                dt=dt,
                config=self.config,
                orth_center_right=orth_center_right,
                is_hermitian=not self.has_lindblad_noise,
            )

            self.state.orthogonality_center = r if orth_center_right else l

    def progress(self) -> None:
        """
        Do one unit of simulation work given the current state.
        Update the state accordingly.
        The state of the simulation is stored in self.sweep_index and self.swipe_direction.
        """
        if self.is_finished():
            return

        delta_time = self.target_time - self.current_time

        assert self.qubit_count >= 1
        if 1 <= self.qubit_count <= 2:
            # Corner case: only 1 or 2 qubits
            assert self.swipe_direction == SwipeDirection.LEFT_TO_RIGHT
            assert self.sweep_index == 0

            if self.qubit_count == 1:
                self._evolve(0, dt=delta_time)
            else:
                self._evolve(0, 1, dt=delta_time, orth_center_right=False)

            self.sweep_complete()

        elif (
            self.sweep_index < self.qubit_count - 2
            and self.swipe_direction == SwipeDirection.LEFT_TO_RIGHT
        ):
            # Left-to-right swipe of TDVP
            self._evolve(
                self.sweep_index,
                self.sweep_index + 1,
                dt=delta_time / 2,
                orth_center_right=True,
            )
            self.left_baths.append(
                new_left_bath(
                    self.get_current_left_bath(),
                    self.state.factors[self.sweep_index],
                    self.hamiltonian.factors[self.sweep_index],
                ).to(self.state.factors[self.sweep_index + 1].device)
            )
            self._evolve(self.sweep_index + 1, dt=-delta_time / 2)
            self.right_baths.pop()
            self.sweep_index += 1

        elif (
            self.sweep_index == self.qubit_count - 2
            and self.swipe_direction == SwipeDirection.LEFT_TO_RIGHT
        ):
            # Time-evolution of the rightmost 2 tensors
            self._evolve(
                self.sweep_index,
                self.sweep_index + 1,
                dt=delta_time,
                orth_center_right=False,
            )
            self.swipe_direction = SwipeDirection.RIGHT_TO_LEFT

        elif (
            1 <= self.sweep_index and self.swipe_direction == SwipeDirection.RIGHT_TO_LEFT
        ):
            # Right-to-left swipe of TDVP
            assert self.sweep_index <= self.qubit_count - 2
            self.right_baths.append(
                new_right_bath(
                    self.get_current_right_bath(),
                    self.state.factors[self.sweep_index + 1],
                    self.hamiltonian.factors[self.sweep_index + 1],
                ).to(self.state.factors[self.sweep_index].device)
            )
            if not self.has_lindblad_noise:
                # Free memory because it won't be used anymore
                deallocate_tensor(self.right_baths[-2])

            self._evolve(self.sweep_index, dt=-delta_time / 2)
            self.left_baths.pop()

            self._evolve(
                self.sweep_index - 1,
                self.sweep_index,
                dt=delta_time / 2,
                orth_center_right=False,
            )
            self.sweep_index -= 1

            if self.sweep_index == 0:
                self.sweep_complete()
                self.swipe_direction = SwipeDirection.LEFT_TO_RIGHT

        else:
            raise Exception("Didn't expect this")

        self.save_simulation()

    def sweep_complete(self) -> None:
        self.current_time = self.target_time
        self.timestep_complete()

    def timestep_complete(self) -> None:
        self.fill_results()
        self.timestep_index += 1
        if self.is_masked and self.current_time >= self.slm_end_time:
            self.is_masked = False
            self.hamiltonian = make_H(
                interaction_matrix=self.full_interaction_matrix,
                hamiltonian_type=self.hamiltonian_type,
                dim=self.dim,
                num_gpus_to_use=self.resolved_num_gpus,
            )

        if not self.is_finished():
            self.target_time = self.target_times[self.timestep_index + 1]
            update_H(
                hamiltonian=self.hamiltonian,
                omega=self.omega[self.timestep_index, :],
                delta=self.delta[self.timestep_index, :],
                phi=self.phi[self.timestep_index, :],
                noise=self.lindblad_noise,
            )
            self.init_baths()

        self.statistics.data.append(time.time() - self.time)
        self.statistics(
            self.config,
            self.current_time / self.target_times[-1],
            self.state,
            self.hamiltonian,
            self.results,
        )
        self.time = time.time()

    def save_simulation(self) -> None:
        if self.last_save_time > time.time() - self.config.autosave_dt:
            return

        basename = self.autosave_file
        with open(basename.with_suffix(".new"), "wb") as file_handle:
            pickle.dump(self, file_handle)
        if basename.is_file():
            os.rename(basename, basename.with_suffix(".bak"))

        os.rename(basename.with_suffix(".new"), basename)
        autosave_filesize = os.path.getsize(self.autosave_file) / 1e6

        if basename.with_suffix(".bak").is_file():
            os.remove(basename.with_suffix(".bak"))

        self.last_save_time = time.time()

        self.config.logger.debug(
            f"Saved simulation state in file {self.autosave_file} ({autosave_filesize}MB)"
        )

    def fill_results(self) -> None:
        normalized_state = 1 / self.state.norm() * self.state

        current_time_int: int = round(self.current_time)
        fractional_time = self.current_time / self.target_times[-1]
        assert abs(self.current_time - current_time_int) < 1e-10

        if self.well_prepared_qubits_filter is None:
            for callback in self.config.observables:
                callback(
                    self.config,
                    fractional_time,
                    normalized_state,
                    self.hamiltonian,
                    self.results,
                )
            return

        full_mpo, full_state = None, None
        for callback in self.config.observables:
            time_tol = 0.5 / self.target_times[-1] + 1e-10
            if (
                callback.evaluation_times is not None
                and self.config.is_time_in_evaluation_times(
                    fractional_time, callback.evaluation_times, tol=time_tol
                )
            ) or self.config.is_evaluation_time(fractional_time, tol=time_tol):

                if full_mpo is None or full_state is None:
                    # Only do this potentially expensive step once and when needed.
                    full_mpo = MPO(
                        extended_mpo_factors(
                            self.hamiltonian.factors, self.well_prepared_qubits_filter
                        )
                    )
                    full_state = MPS(
                        extended_mps_factors(
                            normalized_state.factors,
                            self.well_prepared_qubits_filter,
                        ),
                        num_gpus_to_use=None,  # Keep the already assigned devices.
                        orthogonality_center=get_extended_site_index(
                            self.well_prepared_qubits_filter,
                            normalized_state.orthogonality_center,
                        ),
                        eigenstates=normalized_state.eigenstates,
                    )

                callback(self.config, fractional_time, full_state, full_mpo, self.results)

    def permute_results(self, results: Results, permute: bool) -> Results:
        if permute:
            inv_perm = optimat.inv_permutation(self.qubit_permutation)
            permute_bitstrings(results, inv_perm)
            permute_occupations_and_correlations(results, inv_perm)
            permute_atom_order(results, inv_perm)
        return results


def permute_bitstrings(results: Results, perm: torch.Tensor) -> None:
    if "bitstrings" not in results.get_result_tags():
        return
    uuid_bs = results._find_uuid("bitstrings")

    results._results[uuid_bs] = [
        Counter({optimat.permute_string(bstr, perm): c for bstr, c in bs_counter.items()})
        for bs_counter in results._results[uuid_bs]
    ]


def permute_occupations_and_correlations(results: Results, perm: torch.Tensor) -> None:
    for corr in ["occupation", "correlation_matrix"]:
        if corr not in results.get_result_tags():
            continue

        uuid_corr = results._find_uuid(corr)
        corrs = results._results[uuid_corr]
        results._results[uuid_corr] = (
            [  # vector quantities become lists after results are serialized (e.g. for checkpoints)
                optimat.permute_tensor(
                    corr if isinstance(corr, torch.Tensor) else torch.tensor(corr), perm
                )
                for corr in corrs
            ]
        )


def permute_atom_order(results: Results, perm: torch.Tensor) -> None:
    at_ord = list(results.atom_order)
    at_ord = optimat.permute_list(at_ord, perm)
    results.atom_order = tuple(at_ord)


class NoisyMPSBackendImpl(MPSBackendImpl):
    """
    Version of MPSBackendImpl with non-zero lindbladian noise.
    Implements the Monte-Carlo Wave Function jump method.
    """

    jump_threshold: float
    aggregated_lindblad_ops: torch.Tensor
    norm_gap_before_jump: float
    root_finder: Optional[BrentsRootFinder]

    def __init__(self, config: MPSConfig, pulser_data: PulserData):
        super().__init__(config, pulser_data)
        self.lindblad_ops = pulser_data.lindblad_ops
        self.root_finder = None

        assert self.has_lindblad_noise

    def init_lindblad_noise(self) -> None:
        stacked = torch.stack(self.lindblad_ops)
        # The below is used for batch computation of noise collapse weights.
        self.aggregated_lindblad_ops = stacked.conj().transpose(1, 2) @ stacked

        self.lindblad_noise = compute_noise_from_lindbladians(self.lindblad_ops)

    def set_jump_threshold(self, bound: float) -> None:
        self.jump_threshold = random.uniform(0.0, bound)
        self.norm_gap_before_jump = self.state.norm().item() ** 2 - self.jump_threshold

    def init(self) -> None:
        self.init_lindblad_noise()
        super().init()
        self.set_jump_threshold(1.0)

    def sweep_complete(self) -> None:
        previous_time = self.current_time
        self.current_time = self.target_time
        previous_norm_gap_before_jump = self.norm_gap_before_jump
        self.norm_gap_before_jump = self.state.norm().item() ** 2 - self.jump_threshold

        if self.root_finder is None:
            # No quantum jump location finding in progress
            if self.norm_gap_before_jump < 0:
                # Initiate quantum jump location finding
                # Jump occurs when norm_gap_before_jump ~ 0
                self.root_finder = BrentsRootFinder(
                    start=previous_time,
                    end=self.current_time,
                    f_start=previous_norm_gap_before_jump,
                    f_end=self.norm_gap_before_jump,
                    epsilon=1,
                )
                self.target_time = self.root_finder.get_next_abscissa()
            else:
                self.timestep_complete()

            return

        self.norm_gap_before_jump = self.state.norm().item() ** 2 - self.jump_threshold
        self.root_finder.provide_ordinate(self.current_time, self.norm_gap_before_jump)

        if self.root_finder.is_converged(tolerance=1):
            self.do_random_quantum_jump()
            self.target_time = self.target_times[self.timestep_index + 1]
            self.root_finder = None
        else:
            self.target_time = self.root_finder.get_next_abscissa()

    def do_random_quantum_jump(self) -> None:
        jump_operator_weights = self.state.expect_batch(self.aggregated_lindblad_ops).real
        jumped_qubit_index, jump_operator = random.choices(
            [
                (qubit, op)
                for qubit in range(self.state.num_sites)
                for op in self.lindblad_ops
            ],
            weights=jump_operator_weights.view(-1).tolist(),
        )[0]

        self.state.apply(jumped_qubit_index, jump_operator)
        self.state.orthogonalize(0)
        self.state *= 1 / self.state.norm()
        self.init_baths()

        norm_after_normalizing = self.state.norm().item()
        assert math.isclose(norm_after_normalizing, 1, abs_tol=1e-10)
        self.set_jump_threshold(norm_after_normalizing**2)

    def fill_results(self) -> None:
        # Remove the noise from self.hamiltonian for the callbacks.
        # Since update_H is called at the start of do_time_step this is safe.
        update_H(
            hamiltonian=self.hamiltonian,
            omega=self.omega[self.timestep_index - 1, :],  # Meh
            delta=self.delta[self.timestep_index - 1, :],
            phi=self.phi[self.timestep_index - 1, :],
            noise=torch.zeros(self.dim, self.dim, dtype=dtype),  # no noise
        )

        super().fill_results()


class DMRGBackendImpl(MPSBackendImpl):
    def __init__(
        self,
        mps_config: MPSConfig,
        pulser_data: PulserData,
        energy_tolerance: float = 1e-5,
        max_sweeps: int = 2000,
    ):

        if mps_config.noise_model.noise_types != ():
            raise NotImplementedError(
                "DMRG solver does not currently support noise types"
                f"you are using: {mps_config.noise_model.noise_types}"
            )
        super().__init__(mps_config, pulser_data)
        self.previous_energy: Optional[float] = None
        self.current_energy: Optional[float] = None
        self.sweep_count: int = 0
        self.energy_tolerance: float = energy_tolerance
        self.max_sweeps: int = max_sweeps

    def convergence_check(self, energy_tolerance: float) -> bool:
        if self.previous_energy is None or self.current_energy is None:
            return False
        return abs(self.current_energy - self.previous_energy) < energy_tolerance

    def progress(self) -> None:
        if self.is_finished():
            return

        # perform one two-site energy minimization and update
        idx = self.sweep_index
        assert self.swipe_direction in (
            SwipeDirection.LEFT_TO_RIGHT,
            SwipeDirection.RIGHT_TO_LEFT,
        ), "Unknown Swipe direction"

        orth_center_right = self.swipe_direction == SwipeDirection.LEFT_TO_RIGHT
        new_L, new_R, energy = minimize_energy_pair(
            state_factors=self.state.factors[idx : idx + 2],
            ham_factors=self.hamiltonian.factors[idx : idx + 2],
            baths=(self.left_baths[-1], self.right_baths[-1]),
            orth_center_right=orth_center_right,
            config=self.config,
            residual_tolerance=self.config.precision,
        )
        self.state.factors[idx], self.state.factors[idx + 1] = new_L, new_R
        self.state.orthogonality_center = idx + 1 if orth_center_right else idx
        self.current_energy = energy

        # updating baths and orthogonality center
        if self.swipe_direction == SwipeDirection.LEFT_TO_RIGHT:
            self._left_to_right_update(idx)
        elif self.swipe_direction == SwipeDirection.RIGHT_TO_LEFT:
            self._right_to_left_update(idx)
        else:
            raise Exception("Did not expect this")

        self.save_simulation()

    def _left_to_right_update(self, idx: int) -> None:
        if idx < self.qubit_count - 2:
            self.left_baths.append(
                new_left_bath(
                    self.get_current_left_bath(),
                    self.state.factors[idx],
                    self.hamiltonian.factors[idx],
                ).to(self.state.factors[idx + 1].device)
            )
            self.right_baths.pop()
            self.sweep_index += 1

        if self.sweep_index == self.qubit_count - 2:
            self.swipe_direction = SwipeDirection.RIGHT_TO_LEFT

    def _right_to_left_update(self, idx: int) -> None:
        if idx > 0:
            self.right_baths.append(
                new_right_bath(
                    self.get_current_right_bath(),
                    self.state.factors[idx + 1],
                    self.hamiltonian.factors[idx + 1],
                ).to(self.state.factors[idx].device)
            )
            self.left_baths.pop()
            self.sweep_index -= 1

        if self.sweep_index == 0:
            self.state.orthogonalize(0)
            self.swipe_direction = SwipeDirection.LEFT_TO_RIGHT
            self.sweep_count += 1
            self.sweep_complete()

    def sweep_complete(self) -> None:
        # This marks the end of one full sweep: checking convergence
        if self.convergence_check(self.energy_tolerance):
            self.current_time = self.target_time
            self.timestep_complete()
        elif self.sweep_count + 1 > self.max_sweeps:
            # not converged
            raise RuntimeError(f"DMRG did not converge after {self.max_sweeps} sweeps")
        else:
            # not converged for the current sweep. restart
            self.previous_energy = self.current_energy

        assert self.sweep_index == 0
        assert self.state.orthogonality_center == 0
        assert self.swipe_direction == SwipeDirection.LEFT_TO_RIGHT
        self.current_energy = None


def create_impl(sequence: Sequence, config: MPSConfig) -> MPSBackendImpl:
    pulser_data = PulserData(sequence=sequence, config=config, dt=config.dt)

    if pulser_data.has_lindblad_noise:
        return NoisyMPSBackendImpl(config, pulser_data)
    if config.solver == Solver.DMRG:
        return DMRGBackendImpl(config, pulser_data)
    return MPSBackendImpl(config, pulser_data)



import torch
from typing import Callable, Sequence

from emu_base import krylov_exp
from emu_base.math.krylov_energy_min import krylov_energy_minimization
from emu_base.utils import deallocate_tensor
from emu_mps import MPS, MPO
from emu_mps.utils import split_matrix
from emu_mps.mps_config import MPSConfig


def make_op(
    time_step: float | complex,
    state_factors: Sequence[torch.Tensor],
    baths: tuple[torch.Tensor, torch.Tensor],
    ham_factors: Sequence[torch.Tensor],
) -> tuple[torch.Tensor, torch.device, Callable[[torch.Tensor], torch.Tensor]]:
    assert len(state_factors) == 2
    assert len(baths) == 2
    assert len(ham_factors) == 2

    left_state_factor, right_state_factor = state_factors
    left_bath, right_bath = baths
    left_ham_factor, right_ham_factor = ham_factors

    left_device = left_state_factor.device
    right_device = right_state_factor.device

    left_bond_dim = left_state_factor.shape[0]
    right_bond_dim = right_state_factor.shape[-1]

    # Computation is done on left_device (arbitrary)

    right_state_factor = right_state_factor.to(left_device)

    combined_state_factors = torch.tensordot(
        left_state_factor, right_state_factor, dims=1
    ).reshape(left_bond_dim, 4, right_bond_dim)

    deallocate_tensor(left_state_factor)
    deallocate_tensor(right_state_factor)

    left_ham_factor = left_ham_factor.to(left_device)
    right_ham_factor = right_ham_factor.to(left_device)
    right_bath = right_bath.to(left_device)

    combined_hamiltonian_factors = (
        torch.tensordot(left_ham_factor, right_ham_factor, dims=1)
        .transpose(2, 3)
        .contiguous()
        .view(left_ham_factor.shape[0], 4, 4, -1)
    )

    def op(x: torch.Tensor) -> torch.Tensor:
        return time_step * apply_effective_Hamiltonian(
            x, combined_hamiltonian_factors, left_bath, right_bath
        )

    return combined_state_factors, right_device, op


def new_right_bath(
    bath: torch.Tensor, state: torch.Tensor, op: torch.Tensor
) -> torch.Tensor:
    bath = torch.tensordot(state, bath, ([2], [2]))
    bath = torch.tensordot(op.to(bath.device), bath, ([2, 3], [1, 3]))
    bath = torch.tensordot(state.conj(), bath, ([1, 2], [1, 3]))
    return bath


"""
function to compute the right baths. The three indices in the bath are as follows:
(bond of state conj, bond of operator, bond of state)
The baths have shape
-xx
-xx
-xx
with the index ordering (top, middle, bottom)
bath tensors are put on the device of the factor to the left
"""


def right_baths(state: MPS, op: MPO, final_qubit: int) -> list[torch.Tensor]:
    state_factor = state.factors[-1]
    bath = torch.ones(1, 1, 1, device=state_factor.device, dtype=state_factor.dtype)
    baths = [bath]
    for i in range(len(state.factors) - 1, final_qubit - 1, -1):
        bath = new_right_bath(bath, state.factors[i], op.factors[i])
        bath = bath.to(state.factors[i - 1].device)
        baths.append(bath)
    return baths


"""
Computes H(psi) where
    x-    -x
    x  ||  x             ||
H = x- xx -x  and psi = -xx-
    x  ||  x
    x-    -x

Expects the two qubit factors of the MPS precontracted,
with one 'fat' physical index of dim 4 and index ordering
(left bond, physical index, right bond):
         ||
      -xxxxxx-
The Hamiltonian should have an index ordering of
(left bond, out, in, right bond).
The baths must have shape (top, middle, bottom).
All tensors must be on the same device
"""


def apply_effective_Hamiltonian(
    state: torch.Tensor,
    ham: torch.Tensor,
    left_bath: torch.Tensor,
    right_bath: torch.Tensor,
) -> torch.Tensor:
    assert left_bath.ndim == 3 and left_bath.shape[0] == left_bath.shape[2]
    assert right_bath.ndim == 3 and right_bath.shape[0] == right_bath.shape[2]
    assert left_bath.shape[2] == state.shape[0] and right_bath.shape[2] == state.shape[2]
    assert left_bath.shape[1] == ham.shape[0] and right_bath.shape[1] == ham.shape[3]

    # the optimal contraction order depends on the details
    # this order seems to be pretty balanced, but needs to be
    # revisited when use-cases are more well-known
    state = torch.tensordot(left_bath, state, 1)
    state = state.permute(0, 3, 1, 2)
    ham = ham.permute(0, 2, 1, 3)
    state = state.view(state.shape[0], state.shape[1], -1).contiguous()
    ham = ham.contiguous().view(-1, ham.shape[2], ham.shape[3])
    state = torch.tensordot(state, ham, 1)
    state = state.permute(0, 2, 1, 3)
    state = state.contiguous().view(state.shape[0], state.shape[1], -1)
    right_bath = right_bath.permute(2, 1, 0)
    right_bath = right_bath.contiguous().view(-1, right_bath.shape[2])
    state = torch.tensordot(state, right_bath, 1)
    return state


_TIME_CONVERSION_COEFF = 0.001  # Omega and delta are given in rad/μs, dt in ns


def evolve_pair(
    *,
    state_factors: list[torch.Tensor],
    baths: tuple[torch.Tensor, torch.Tensor],
    ham_factors: list[torch.Tensor],
    dt: float,
    orth_center_right: bool,
    is_hermitian: bool,
    config: MPSConfig,
) -> tuple[torch.Tensor, torch.Tensor]:
    """
    Time evolution of a pair of tensors of a tensor train using baths and truncated SVD.
    Returned state tensors are kept on their respective devices.

    The input state tensor objects become invalid after calling that function.
    """

    time_step = -1j * _TIME_CONVERSION_COEFF * dt
    combined_state_factors, right_device, op = make_op(
        time_step=time_step,
        state_factors=state_factors,
        baths=baths,
        ham_factors=ham_factors,
    )
    left_bond_dim = combined_state_factors.shape[0]
    right_bond_dim = combined_state_factors.shape[-1]

    evol = krylov_exp(
        op,
        combined_state_factors,
        exp_tolerance=config.precision * config.extra_krylov_tolerance,
        norm_tolerance=config.precision * config.extra_krylov_tolerance,
        max_krylov_dim=config.max_krylov_dim,
        is_hermitian=is_hermitian,
    ).view(left_bond_dim * 2, 2 * right_bond_dim)

    l, r = split_matrix(
        evol,
        max_error=config.precision,
        max_rank=config.max_bond_dim,
        orth_center_right=orth_center_right,
        preserve_norm=not is_hermitian,  # only relevant for computing jump times
    )

    return l.view(left_bond_dim, 2, -1), r.view(-1, 2, right_bond_dim).to(right_device)


def evolve_single(
    *,
    state_factor: torch.Tensor,
    baths: tuple[torch.Tensor, torch.Tensor],
    ham_factor: torch.Tensor,
    dt: float,
    is_hermitian: bool,
    config: MPSConfig,
) -> torch.Tensor:
    """
    Time evolution of a single tensor of a tensor train using baths.

    The input state tensor object becomes invalid after calling that function.
    """
    assert len(baths) == 2

    left_bath, right_bath = baths

    def op(x: torch.Tensor) -> torch.Tensor:
        return (
            -_TIME_CONVERSION_COEFF
            * 1j
            * dt
            * apply_effective_Hamiltonian(
                x,
                ham_factor,
                left_bath,
                right_bath,
            )
        )

    return krylov_exp(
        op,
        state_factor,
        exp_tolerance=config.precision * config.extra_krylov_tolerance,
        norm_tolerance=config.precision * config.extra_krylov_tolerance,
        max_krylov_dim=config.max_krylov_dim,
        is_hermitian=is_hermitian,
    )


def minimize_energy_pair(
    *,
    state_factors: Sequence[torch.Tensor],
    baths: tuple[torch.Tensor, torch.Tensor],
    ham_factors: Sequence[torch.Tensor],
    orth_center_right: bool,
    config: MPSConfig,
    residual_tolerance: float,
) -> tuple[torch.Tensor, torch.Tensor, float]:
    """
    Minimizes the state factors (ψ_i, ψ_{i+1}) using the Lanczos/Arnoldi method
    """

    time_step = 1.0
    combined_state_factors, right_device, op = make_op(
        time_step=time_step,
        state_factors=state_factors,
        baths=baths,
        ham_factors=ham_factors,
    )

    left_bond_dim = combined_state_factors.shape[0]
    right_bond_dim = combined_state_factors.shape[-1]

    updated_state, updated_energy = krylov_energy_minimization(
        op,
        combined_state_factors,
        norm_tolerance=config.precision * config.extra_krylov_tolerance,
        residual_tolerance=residual_tolerance,
        max_krylov_dim=config.max_krylov_dim,
    )
    updated_state = updated_state.view(left_bond_dim * 2, 2 * right_bond_dim)

    l, r = split_matrix(
        updated_state,
        max_error=config.precision,
        max_rank=config.max_bond_dim,
        orth_center_right=orth_center_right,
    )

    return (
        l.view(left_bond_dim, 2, -1),
        r.view(-1, 2, right_bond_dim).to(right_device),
        updated_energy,
    )



from pulser.backend import (
    BitStrings,
    CorrelationMatrix,
    Energy,
    EnergyVariance,
    Expectation,
    Fidelity,
    Occupation,
    StateResult,
    EnergySecondMoment,
)
from .mps_config import MPSConfig, Solver
from .mpo import MPO
from .mps import MPS, inner
from .mps_backend import MPSBackend
from .observables import EntanglementEntropy


__all__ = [
    "__version__",
    "MPO",
    "MPS",
    "inner",
    "MPSConfig",
    "Solver",
    "MPSBackend",
    "StateResult",
    "BitStrings",
    "Occupation",
    "CorrelationMatrix",
    "Expectation",
    "Fidelity",
    "Energy",
    "EnergyVariance",
    "EnergySecondMoment",
    "EntanglementEntropy",
]

__version__ = "2.4.4"



from typing import List, Optional
import torch


def new_left_bath(
    bath: torch.Tensor, state: torch.Tensor, op: torch.Tensor
) -> torch.Tensor:
    # this order is more efficient than contracting the op first in general
    bath = torch.tensordot(bath, state.conj(), ([0], [0]))
    bath = torch.tensordot(bath, op.to(bath.device), ([0, 2], [0, 1]))
    bath = torch.tensordot(bath, state, ([0, 2], [0, 1]))
    return bath


def _determine_cutoff_index(d: torch.Tensor, max_error: float) -> int:
    assert max_error > 0
    squared_max_error = max_error * max_error
    acc = 0.0
    for i in range(d.shape[0]):
        acc += d[i].item()
        if acc > squared_max_error:
            return i
    return 0  # type: ignore[no-any-return]


def split_matrix(
    m: torch.Tensor,
    max_error: float = 1e-5,
    max_rank: int = 1024,
    orth_center_right: bool = True,
    preserve_norm: bool = False,
) -> tuple[torch.Tensor, torch.Tensor]:
    """
    Computes a low-rank approximation split of m using the Eckart-Young-Mirsky theorem.
    """
    assert m.ndim == 2

    if orth_center_right:
        d, q = torch.linalg.eigh(m @ m.T.conj())
        max_bond = max(
            _determine_cutoff_index(d, max_error),
            d.shape[0] - max_rank,
        )
        left = q[:, max_bond:]
        right = left.T.conj() @ m
        if preserve_norm:
            old_norm2 = torch.sum(d)
            new_norm2 = torch.sum(d[max_bond:])
            right *= torch.sqrt(old_norm2 / new_norm2)
    else:
        d, q = torch.linalg.eigh(m.T.conj() @ m)
        max_bond = max(
            _determine_cutoff_index(d, max_error),
            d.shape[0] - max_rank,
        )
        right = q[:, max_bond:].T.conj_physical()
        left = m @ q[:, max_bond:]
        if preserve_norm:
            old_norm2 = torch.sum(d)
            new_norm2 = torch.sum(d[max_bond:])
            left *= torch.sqrt(old_norm2 / new_norm2)

    return left, right


def truncate_impl(
    factors: list[torch.Tensor], precision: float, max_bond_dim: int
) -> None:
    """
    Eigenvalues-based truncation of a matrix product.
    An in-place operation.

    Note:
        Sweeps from right to left.
        Requires the matrix product to be orthogonalized on the last element.
        At each step moves the orthogonal center to the left while truncating.
    """
    for i in range(len(factors) - 1, 0, -1):
        factor_shape = factors[i].shape

        l, r = split_matrix(
            factors[i].view(factor_shape[0], -1),
            max_error=precision,
            max_rank=max_bond_dim,
            orth_center_right=False,
        )

        factors[i] = r.view(-1, *factor_shape[1:])
        factors[i - 1] = torch.tensordot(
            factors[i - 1], l.to(factors[i - 1].device), dims=1
        )


def assign_devices(tensors: List[torch.Tensor], num_gpus_to_use: int) -> None:
    """
    Evenly distributes each tensor in the list to a device.
    If num_gpus_to_use is 0, then all tensors go to CPU.
    """
    num_gpus_to_use = min(len(tensors), num_gpus_to_use)

    if num_gpus_to_use <= 0:
        for i in range(len(tensors)):
            tensors[i] = tensors[i].to("cpu")
        return

    tensors_per_device = len(tensors) // num_gpus_to_use

    if len(tensors) % num_gpus_to_use != 0:
        tensors_per_device += 1

    for i in range(len(tensors)):
        tensors[i] = tensors[i].to(f"cuda:{i // tensors_per_device}")


def extended_mps_factors(
    mps_factors: list[torch.Tensor], where: torch.Tensor
) -> list[torch.Tensor]:
    """
    Given a valid list of MPS factors, accounting for qubits marked as `True` in `where`,
    fills the `False` positions with new qubits in the |0> state.
    """
    assert len(mps_factors) == sum(1 for b in where if b)

    bond_dimension = 1
    factor_index = 0
    result = []
    for is_factor in where:
        assert 0 <= factor_index <= len(mps_factors)

        if is_factor:
            result.append(mps_factors[factor_index])
            bond_dimension = mps_factors[factor_index].shape[2]
            factor_index += 1
        elif factor_index == len(mps_factors):
            factor = torch.zeros(
                bond_dimension, 2, 1, dtype=torch.complex128
            )  # FIXME: assign device
            factor[:, 0, :] = torch.eye(bond_dimension, 1)
            bond_dimension = 1
            result.append(factor)
        else:
            factor = torch.zeros(
                bond_dimension,
                2,
                bond_dimension,
                dtype=torch.complex128,  # FIXME: assign device
            )
            factor[:, 0, :] = torch.eye(bond_dimension, bond_dimension)
            result.append(factor)
    return result


def extended_mpo_factors(
    mpo_factors: list[torch.Tensor], where: torch.Tensor
) -> list[torch.Tensor]:
    """
    Given a valid list of MPO factors, accounting for qubits marked as `True` in `where`,
    fills the `False` positions with new MPO identity factors.
    """
    assert len(mpo_factors) == sum(1 for b in where if b)

    bond_dimension = 1
    factor_index = 0
    result = []
    for is_factor in where:
        assert 0 <= factor_index <= len(mpo_factors)

        if is_factor:
            result.append(mpo_factors[factor_index])
            bond_dimension = mpo_factors[factor_index].shape[3]
            factor_index += 1
        elif factor_index == len(mpo_factors):
            factor = torch.zeros(bond_dimension, 2, 2, 1, dtype=torch.complex128)
            factor[:, 0, 0, :] = torch.eye(bond_dimension, 1)
            factor[:, 1, 1, :] = torch.eye(bond_dimension, 1)
            bond_dimension = 1
            result.append(factor)
        else:
            factor = torch.zeros(
                bond_dimension, 2, 2, bond_dimension, dtype=torch.complex128
            )
            factor[:, 0, 0, :] = torch.eye(bond_dimension, bond_dimension)
            factor[:, 1, 1, :] = torch.eye(bond_dimension, bond_dimension)
            result.append(factor)
    return result


def get_extended_site_index(
    where: torch.Tensor, desired_index: Optional[int]
) -> Optional[int]:
    """
    Returns the index in `where` that has `desired_index` preceding True elements.

    This function is used to find the index of the orthogonality center in an MPS obtained
    with `extended_mps_factors` in the presence of dark qubits:
        `where` is the mask specifying whether qubits are well-prepared.
        `desired_index` is the index of the orthogonality center of the MPS without dark qubits.
        The return value is then the index of the orthogonality center
        in the full MPS with added dark qubits.
    """

    if desired_index is None:
        return None

    index = -1
    for extended_index, boolean_value in enumerate(where):
        if boolean_value:
            index += 1
            if index == desired_index:
                return extended_index

    raise ValueError(f"Index {desired_index} does not exist")


def tensor_trace(tensor: torch.Tensor, dim1: int, dim2: int) -> torch.Tensor:
    """
    Contract two legs of a single tensor.
    """
    assert tensor.shape[dim1] == tensor.shape[dim2], "dimensions should match"
    return tensor.diagonal(offset=0, dim1=dim1, dim2=dim2).sum(-1)



import itertools

import torch
from scipy.sparse import csr_matrix
from scipy.sparse.csgraph import reverse_cuthill_mckee

from emu_mps.optimatrix.permutations import permute_tensor


def is_symmetric(matrix: torch.Tensor, tol: float = 1e-8) -> bool:
    return torch.allclose(matrix, matrix.T, atol=tol)


def matrix_bandwidth(mat: torch.Tensor) -> float:
    """matrix_bandwidth(matrix: torch.tensor) -> torch.Tensor

    Computes bandwidth as max weighted distance between columns of
    a square matrix as `max (abs(matrix[i, j] * (j - i))`.

             abs(j-i)
          |<--------->|
        (i,i)       (i,j)
          |           |
    | *   .   .   .   .   . |
    | .   *   .   .   a   . |
    | .   .   *   .   .   . |
    | .   .   .   *   .   . |
    | .   .   .   .   *   . |
    | .   .   .   .   .   * |

    Distance from the main diagonal `[i,i]` and element `m[i,j]` along row is
    `abs(j-i)` and therefore the weighted distance is `abs(matrix[i, j] * (j - i))`

    Parameters
    -------
    matrix :
        square matrix nxn

    Returns
    -------
        bandwidth of the input matrix

    Example:
    -------
    >>> matrix = torch.tensor([
    ...     [1.0, -17.0, 2.4],
    ...     [9.0, 1.0, -10.0],
    ...     [-15.0, 20.0, 1.0]
    ... ])
    >>> matrix_bandwidth(matrix)  # because abs(-15 * (0 - 2)) = 30.0
    30.0
    """

    n = mat.shape[0]

    i_arr = torch.arange(n).view(-1, 1)  # shape (n, 1)
    j_arr = torch.arange(n).view(1, -1)  # shape (1, n)

    weighted = torch.abs(mat * (j_arr - i_arr))
    return torch.max(weighted).to(mat.dtype).item()


def minimize_bandwidth_above_threshold(
    mat: torch.Tensor, threshold: float
) -> torch.Tensor:
    """
    minimize_bandwidth_above_threshold(matrix, trunc) -> permutation_lists

    Finds a permutation list that minimizes a bandwidth of a symmetric matrix `A = A.T`
    using the reverse Cuthill-Mckee algorithm from `scipy.sparse.csgraph.reverse_cuthill_mckee`.
    Matrix elements below a threshold `m[i,j] < threshold` are considered as 0.

    Parameters
    -------
    matrix :
        symmetric square matrix
    threshold :
        matrix elements `m[i,j] < threshold` are considered as 0

    Returns
    -------
        permutation list that minimizes matrix bandwidth for a given threshold

    Example:
    -------
    >>> matrix = torch.tensor([
    ...     [1, 2, 3],
    ...     [2, 5, 6],
    ...     [3, 6, 9]
    ... ], dtype=torch.float32)
    >>> threshold = 3
    >>> minimize_bandwidth_above_threshold(matrix, threshold)
    tensor([1, 2, 0], dtype=torch.int32)
    """

    m_trunc = mat.clone()
    m_trunc[mat < threshold] = 0.0

    matrix_np = csr_matrix(m_trunc.numpy())  # SciPy's RCM compatibility
    rcm_perm = reverse_cuthill_mckee(matrix_np, symmetric_mode=True)
    return torch.from_numpy(rcm_perm.copy())  # translation requires copy


def minimize_bandwidth_global(mat: torch.Tensor) -> torch.Tensor:
    """
    minimize_bandwidth_global(matrix) -> list

    Does one optimisation step towards finding
    a permutation of a matrix that minimizes matrix bandwidth.

    Parameters
    -------
    matrix :
        symmetric square matrix

    Returns
    -------
        permutation order that minimizes matrix bandwidth

    Example
    -------
    >>> matrix = torch.tensor([
    ...     [1, 2, 3],
    ...     [2, 5, 6],
    ...     [3, 6, 9]
    ... ], dtype=torch.float32)
    >>> minimize_bandwidth_global(matrix)
    tensor([2, 1, 0], dtype=torch.int32)
    """
    mat_amplitude = torch.max(torch.abs(mat))

    permutations = (
        minimize_bandwidth_above_threshold(mat, trunc.item() * mat_amplitude)
        for trunc in torch.arange(0.1, 1.0, 0.01)
    )

    opt_permutation = min(
        permutations, key=lambda perm: matrix_bandwidth(permute_tensor(mat, perm))
    )

    return opt_permutation


def minimize_bandwidth_impl(
    matrix: torch.Tensor, initial_perm: torch.Tensor
) -> tuple[torch.Tensor, float]:
    """
    minimize_bandwidth_impl(matrix, initial_perm) -> (optimal_perm, bandwidth)

    Applies initial_perm to a matrix and
    finds the permutation list for a symmetric matrix
    that iteratively minimizes matrix bandwidth.

    Parameters
    -------
    matrix :
        symmetric square matrix
    initial_perm: torch list of integers


    Returns
    -------
        optimal permutation and optimal matrix bandwidth

    Example:
    -------
    Periodic 1D chain
    >>> matrix = torch.tensor([
    ...    [0, 1, 0, 0, 1],
    ...    [1, 0, 1, 0, 0],
    ...    [0, 1, 0, 1, 0],
    ...    [0, 0, 1, 0, 1],
    ...    [1, 0, 0, 1, 0]], dtype=torch.float32)
    >>> id_perm = torch.arange(matrix.shape[0])
    >>> minimize_bandwidth_impl(matrix, id_perm) # [3, 2, 4, 1, 0] does zig-zag
    (tensor([3, 2, 4, 1, 0]), 2.0)

    Simple 1D chain. Cannot be optimised further
    >>> matrix = torch.tensor([
    ...    [0, 1, 0, 0, 0],
    ...    [1, 0, 1, 0, 0],
    ...    [0, 1, 0, 1, 0],
    ...    [0, 0, 1, 0, 1],
    ...    [0, 0, 0, 1, 0]], dtype=torch.float32)
    >>> id_perm = torch.arange(matrix.shape[0])
    >>> minimize_bandwidth_impl(matrix, id_perm)
    (tensor([0, 1, 2, 3, 4]), 1.0)
    """
    L = matrix.shape[0]
    if not torch.equal(initial_perm, torch.arange(L)):
        matrix = permute_tensor(matrix, initial_perm)
    bandwidth = matrix_bandwidth(matrix)
    acc_permutation = initial_perm

    for counter in range(101):
        if counter == 100:
            raise (
                NotImplementedError(
                    "The algorithm takes too many steps, " "probably not converging."
                )
            )

        optimal_perm = minimize_bandwidth_global(matrix)
        test_mat = permute_tensor(matrix, optimal_perm)
        new_bandwidth = matrix_bandwidth(test_mat)

        if bandwidth <= new_bandwidth:
            break

        matrix = test_mat
        acc_permutation = permute_tensor(acc_permutation, optimal_perm)
        bandwidth = new_bandwidth

    return acc_permutation, bandwidth


def minimize_bandwidth(input_matrix: torch.Tensor, samples: int = 100) -> torch.Tensor:
    assert is_symmetric(input_matrix), "Input matrix is not symmetric"
    input_mat = torch.abs(input_matrix)
    # We are interested in strength of the interaction, not sign

    L = input_mat.shape[0]
    rnd_permutations = itertools.chain(
        [torch.arange(L)],  # identity permutation
        [torch.randperm(L) for _ in range(samples)],  # list of random permutations
    )

    opt_permutations_and_opt_bandwidth = (
        minimize_bandwidth_impl(input_mat, rnd_perm) for rnd_perm in rnd_permutations
    )

    best_perm, best_bandwidth = min(
        opt_permutations_and_opt_bandwidth,
        key=lambda perm_and_bandwidth: perm_and_bandwidth[1],
    )

    assert best_bandwidth <= matrix_bandwidth(input_matrix), "Matrix is not optimised"
    return best_perm


if __name__ == "__main__":
    import doctest

    doctest.testmod()
